import torch.nn as nn, numpy as np, matplotlib.pyplot as plt
import torch
from .initializers import initializers_dict


loss_dict = {'mse':nn.MSELoss()}

##### functions to implement
# fit, predict, compile (necessary?), save_weights, load_weights

class Sequential():
    """
    Sequential model class.
    """

    def __init__(self):
        """
        init function
        """
        self.model = nn.Sequential()
        self.last_layers_units = 0
        self.layer_iter = 0
        self.is_compiled = False

    def add(self,layer, input_shape = None, layer_name  = None,input_features = None):
        "adds a layer to the model"

        # if first layer, define input shape
        if self.layer_iter == 0:

            if input_shape == None and input_features == None:
                # ValueError should not be used
                raise ValueError("Please Specify input_shape or input_features for first layer!")

            elif input_features == None: # if input_shape given (if both given , shape will be given preference)
                self.last_layers_units = input_shape[0]

            else: # if input_features_given
                self.last_layers_units = input_features

        # if name not specifed, specify your own name which is the iteration number of layer
        if layer_name == None:
            layer_name = str(self.layer_iter)

        # extract activation and layer params
        layer_type, current_layer_param, activation = layer[0], layer[1], layer[2]

        # currently only for Dense and Dropout!!!!
        if layer_type == 'Dense':
            nn_layer = nn.Linear(self.last_layers_units, current_layer_param)
            self.last_layers_units = current_layer_param
        elif layer_type =='Dropout':
            nn_layer = nn.Dropout(current_layer_param)

        self.model.add_module(layer_name,nn_layer)

        # if activation is None, it will be treated as linear activation and no module will be added for activation!
        if activation != None:
            self.model.add_module(layer_name+'_activation',activation)

        # updating
        self.layer_iter = self.layer_iter + 1


    def compile(self,loss,optimizer):
        """
        compiles the model with given loss function and optimizer
        :param loss: given loss function (str)
        :param optimizer: given optimizer (k_torch.optimizers)
        :return:
        """
        self.optimizer = optimizer[0](self.model.parameters(),lr = optimizer[1],weight_decay = optimizer[2])
        self.loss = loss_dict[loss]
        self.is_compiled = True

    def fit(self,X_train,y_train,epochs,verbose = False,validation_data = None, should_plot_history = False):
        """
        trains the model based on given data. Will show progress based on given parameters

        :param X_train:
        :param y_train:
        :param nb_epochs:
        :param verbose:
        :param validation_data:
        :param should_plot_history:
        :return:
        """

        if self.is_compiled == False:
            raise EnvironmentError("Model is not compiled! Please compile first") # change this value error

        # converting training params to np.float64 dtype
        X_train = X_train.astype(np.float64)
        y_train = y_train.astype(np.float64)

        # converting to Tensors
        X_train, y_train = map(torch.Tensor,[X_train,y_train])

        criterion = self.loss

        if validation_data != None:
            X_test = validation_data[0]
            y_test = validation_data[1]

            # converting testing params to np.float64 dtype
            X_test = X_test.astype(np.float64)
            y_test = y_test.astype(np.float64)

            X_test, y_test = map(torch.Tensor, [X_test, y_test])

        training_losses, val_losses = [], []

        # training
        for epoch in range(epochs):
            # Forward Propagation for training loss
            y_pred = self.model(X_train)

            train_loss = criterion(y_pred, y_train)
            training_losses.append(train_loss)

            if validation_data != None:
                y_val = self.model(X_test)  # for validation
                val_loss = criterion(y_val, y_test)
                val_losses.append(val_loss)

            # print verbose if selected
            if verbose:
                if validation_data == None:
                    print('epoch: ', epoch, 'train_loss: ', train_loss.item())
                else:
                    print('epoch: ', epoch, 'train_loss: ', train_loss.item(), 'val_loss: ', val_loss.item())

            # Zero the gradients
            self.optimizer.zero_grad()

            # perform a backward pass (backpropagation)
            train_loss.backward()

            # Update the parameters
            self.optimizer.step()

        if should_plot_history:
            plt.plot(training_losses, label='training_loss')
            if validation_data!=None:
                plt.plot(val_losses, label='validation_loss')
            plt.ylabel('loss')
            plt.xlabel('number of epochs')
            plt.title('Loss History Visualization')
            plt.legend()
            plt.show()


    def predict(self,X_test):
        """
        makes predictions
        :param X_test:
        :return:
        """

        # converting X_train and y_train to np.float64 dtype
        X_test = X_test.astype(np.float64)

        X_test = torch.Tensor(X_test)
        predictions = self.model(X_test).detach().numpy()

        return predictions

    def save_weights(self,path):
        """
        saves weights to a .dat file
        """
        torch.save(self.model.state_dict(), path)

    def load_weights(self,path):
        """
        loads weights from a .dat file
        """
        self.model.load_state_dict(torch.load(path))

    def initialize_weights(self, initialization = 'xavier'):
        """
        initializes weights with the given initialization
        """
        self.model.apply(initializers_dict[initialization])



import torch.optim as optim

# optimizers
optimizers_dict = {'Adadelta': optim.Adadelta, 'Adagrad': optim.Adagrad, 'Adam': optim.Adam}


def Adadelta(lr = 1, weight_decay = 0):
    return optimizers_dict['Adadelta'], lr, weight_decay

def Adagrad(lr = 0.01, weight_decay = 0):
    return optimizers_dict['Adagrad'], lr, weight_decay

def Adam(lr = 0.001, weight_decay = 0):
    return optimizers_dict['Adam'],lr, weight_decay


from k_torch.layers import Dense
from k_torch.models import Sequential
from k_torch.optimizers import Adam

import numpy as np


# data preparation
X_train = [[3,4],[2,3],[4,3],[2,3],[3,4],[2,3],[7,8],[8,9],[9,9]]
y_train = [1.2,1.3,1.4,1.5,1.6,1.7,7,8,8]
X_train = np.asarray (X_train)
y_train = np.asarray(y_train)
y_train = y_train.reshape(-1,1)

X_val = [[9,7],[1,3]]
y_val = [7.8,1.6]
X_val = np.asarray (X_val)
y_val = np.asarray(y_val)
y_val = y_val.reshape(-1,1)

model = Sequential()
model.add(Dense(2), input_shape = (2,))
model.add(Dense(5))
model.add(Dense(1))

model.compile(loss ='mse', optimizer=Adam(lr = 0.01))

model.fit(X_train,y_train,nb_epochs=500,verbose=False,should_plot_history=False,validation_data={'X_test' : X_val, 'y_test':y_val})




# K torch init file

### version 0.0.7 : 14-05-19
# Fixed error in validation set given as input in fit method

### version 0.0.6 : 13-05-19
# Fixed error in predict function. Converted X_test to np.float type


### version 0.0.4 : 10-05-19
# Fixed layer returning 3 params

### version 0.0.3 : 10-05-19
# changed nb_epochs to epochs in models.fit
# changed validation_data from dict to list in models.fit
# added type casting to np.float64 to X and y in models.fit

name = 'k_torch'


import torch.nn as nn



# for xavier initialization
# initialization function, first checks the module type,
# then applies the desired changes to the weights
def xavier_init(m):
    if type(m) == nn.Linear:
        nn.init.xavier_uniform_(m.weight)



initializers_dict = {'xavier' : xavier_init}




import torch.nn as nn

activation_dict = {'relu' : nn.ReLU(), 'linear' : None}

def Dense(no_of_neurons,activation= 'linear'):
    return 'Dense',no_of_neurons, activation_dict[activation]

def Dropout(p):
    return 'Dropout', p ,None





